from itertools import combinations
from typing import Dict, List, Tuple

import numpy as np
import torch
from torch.utils.data import DataLoader

from .datasets import EEGDataset, EEGDatasetBase


class EEGDataloader:
    def __init__(self, dataset: EEGDataset, batch_size: int):
        self.dataset = dataset
        self.batch_size = batch_size

    def loto(self):
        pass

    @staticmethod
    def _concat_data(
        data: List[Tuple[Dict[int, np.ndarray], Dict[int, int]]]
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        concatenate data from several subjects and make one tensor

        Args:
            data: list of subjects data and label. each value is a dictionary with session and data pairs
            we just need values for training and not session ids

        Returns:
            x,y (torch.Tensor):
        """
        x = np.concatenate([v for d in data for v in d[0].values()], axis=0)
        x = torch.from_numpy(x).float()

        # TODO: refactor
        y = []
        for i in data:
            for k, v in i[1].items():
                y.extend(list(np.repeat(v, i[0][k].shape[0])))
        y = torch.tensor(y)

        return x, y

    def _get_dataloader(
        self, data: torch.Tensor, label: torch.Tensor, shuffle: bool = True
    ) -> DataLoader:
        """
        Returns a DataLoader object that can be used for iterating through batches of data and labels.

        Args:
            data (List[torch.Tensor]): List of input data tensors.
            label (List[torch.Tensor]): List of label tensors.
            shuffle (bool): Whether to shuffle the data before batching. Defaults to True.

        Returns:
            loader: A DataLoader object that can be used to iterate through batches of data and labels.
        """
        dataset = EEGDatasetBase(data, label)
        loader = DataLoader(
            dataset=dataset,
            batch_size=self.batch_size,
            shuffle=shuffle,
            pin_memory=True,
        )
        return loader

    def loso(self, subject_out_num: int = 1) -> Tuple[DataLoader, DataLoader]:
        subject_ids = self.dataset.__get_subject_ids__()
        test_ids_combination = list(combinations(subject_ids, subject_out_num))

        for test_subject_ids in test_ids_combination:
            train_subject_ids = set(subject_ids) - set(test_subject_ids)
            print(
                f"\nPreparing data for\ntrain subjects: {train_subject_ids}\ntest subjects: {test_subject_ids}"
            )

            train_data = [self.dataset.__get_subject__(i) for i in train_subject_ids]
            train_data, train_label = EEGDataloader._concat_data(train_data)

            test_data = [self.dataset.__get_subject__(i) for i in test_subject_ids]
            test_data, test_label = EEGDataloader._concat_data(test_data)

            train_dataloader = self._get_dataloader(train_data, train_label)
            test_dataloader = self._get_dataloader(test_data, test_label)
            yield train_dataloader, test_dataloader
